import enum
import numpy as np
import warnings

import uncertainties as unc
import uncertainties.unumpy as unp
from smpl import debug
from smpl import functions
from smpl import stat
from smpl import util
from smpl import wrap
from smpl import doc
from smpl import data
from numpy.linalg import LinAlgError, inv
from tqdm import tqdm
import enum

from .minuit import _fit_minuit_leastsquares
from .scipy import _fit_curvefit, _fit_odr


unv = unp.nominal_values
usd = unp.std_devs


class Fitter(enum.Enum):
    """
    Different implementations to perform a fit.
    """
    AUTO = 0
    SCIPY_CURVEFIT = 1
    SCIPY_ODR = 2
    MINUIT_LEASTSQUARES = 3


default = {
    'params': [None, "Initial fit parameters", ],
    # 'frange': [None, "Limit the fit to given range. First integer is the lowest and second the highest index.", ],
    # 'fselector': [None, "Function that takes ``x`` and ``y`` as parameters and returns an array mask in order to limit the data points for fitting. Alternatively a mask for selecting elements from datax and datay.", ],
    'fixed_params': [True, "Enable fixing parameters by choosing the same-named variables from ``kwargs``.", ],
    # 'sortbyx': [True, "Enable sorting the x and y data so that x is sorted.", ],
    'maxfev': [10000, "Maximum function evaluations during fitting.", ],
    'epsfcn': [0.0001, "Suitable step length for jacobian approximation.", ],
    'xvar': [None, "Variable in fit function parameters that corresponds to the x axis. If it is None the last of the alphabetical sorted parameters is used.", ],
    # 'bins': [0, "Number of bins for histogram", ],
    # 'binunc': [stat.poisson_dist, "Number of bins for histogram", ],
    'autotqdm': [True, "Auto fitting display tqdm", ],
    # 'xerror': [True, "enable xerrors"],
    # 'yerror': [True, "enable yerrors"],
    'fitter': [Fitter.AUTO, "Choose from :class:`Fitter`s."]
}

# @doc.insert_str("\tDefault kwargs\n\n\t")


@doc.append_doc(data.data_kwargs)
@doc.append_str("\t")
@doc.append_str(doc.table(default, init=False))
@doc.append_str(doc.table({"fit_kwargs": ["default", "description"]}, bottom=False))
def fit_kwargs(kwargs):
    """Set default fit_kwargs if not set.

    """
    kwargs = data.data_kwargs(kwargs)
    for k, v in default.items():
        if not k in kwargs:
            kwargs[k] = v[0]
    return kwargs


# @append_doc(default_kwargs)
def auto(datax, datay, funcs=None, **kwargs):
    """
    Automatically loop over functions and fit the best one.

    Parameters
    ----------
    funcs : function array
        functions to consider as fit. Default all ``smpl.functions``.
    **kwargs : optional
        see :func:`fit_kwargs`.

    Returns
    -------
    The best fit function and it's parameters and a ``lambda`` where the parameters are already applied to the function.

    """
    kwargs = fit_kwargs(kwargs)
    min_sq = None
    best_f = None
    best_ff = None

    if funcs is None:
        funcs = functions.__dict__.values()
    for f in tqdm(funcs, disable=not kwargs['autotqdm']):
        if callable(f):
            try:
                ff = fit(datax, datay, f, **kwargs)
                fy = f(datax, *ff)
            except (ValueError, LinAlgError) as ve:
                debug.msg(ve)
                continue
            sum_sq = np.sum((fy - datay)**2) + np.sum((fy + usd(fy) -
                                                       datay)**2) + np.sum((fy - usd(fy) - datay)**2)
            if min_sq is None or sum_sq < min_sq:
                min_sq = sum_sq
                best_f = f
                best_ff = ff
    # if not best_f is None:
    #    fit(datax,datay,best_f,**kwargs)
    return best_f, best_ff, lambda x: best_f(x, *best_ff)


def fit(datax, datay, function, **kwargs):
    """
    Returns a fit of ``function`` to ``datax`` and ``datay``.

    Parameters
    ----------
    datax : array_like
        X data either as ``unp.uarray`` or ``np.array`` or ``list``
    datay : array_like
        Y data either as ``unp.uarray`` or ``np.array`` or ``list``
    function : func
        Fit function with parameters: ``x``, ``params``
    **kwargs : optional
        see :func:`fit_kwargs`.

    """
    kwargs = fit_kwargs(kwargs)
    x, y, xerr, yerr = fit_split(datax, datay, **kwargs)
    params = None
    if util.has('params', kwargs):
        params = kwargs['params']

    fixed = {}
    vnames = wrap.get_varnames(function, kwargs['xvar'])
    Ntot = len(vnames)-1
    if util.has("fixed_params", kwargs) and kwargs['fixed_params']:
        for i in range(1, len(vnames)):
            if util.has(vnames[i], kwargs):
                fixed[i] = kwargs[vnames[i]]
    # Count parameters for function
    if params is None:
        N = len(vnames)
        params = [1 for i in range(N-1)]
    tmp_params = []
    for i, pi in enumerate(params):
        if not util.has(i+1, fixed):
            tmp_params += [pi]
    params = tmp_params
    N = len(params)

    def tmp(*x):
        tmp_x = []
        j = 1
        # print(x)
        for i in range(1, Ntot+1):
            # print(i," ",j)
            if not util.has(i, fixed):
                tmp_x += [x[j]]
                # print(x[j])
                j = j+1
            else:
                tmp_x += [fixed[i]]

        # print(Ntot)
        # print(tmp_x)
        return unv(wrap.get_lambda(function, kwargs['xvar'])(x[0], *tmp_x))
    fitter = kwargs["fitter"]
    if fitter is Fitter.AUTO:
        if xerr is not None:
            fitter = Fitter.SCIPY_ODR
        else:
            fitter = Fitter.SCIPY_CURVEFIT
    if fitter is Fitter.MINUIT_LEASTSQUARES:
        fit = _fit_minuit_leastsquares(x, y, tmp, params=params, yerr=yerr)
    elif fitter is Fitter.SCIPY_CURVEFIT:
        fit = _fit_curvefit(x, y, tmp, params=params, yerr=yerr)
    elif fitter is Fitter.SCIPY_ODR:
        fit = _fit_odr(x, y, tmp, params=params, xerr=xerr, yerr=yerr)

    rfit = []
    j = 0
    for i in range(1, Ntot+1):
        if not util.has(i, fixed):
            rfit += [fit[j]]
            j = j+1
        else:
            rfit += [fixed[i]]

    return rfit


@doc.insert_doc(stat.Chi2)
def Chi2(datax, datay, function, ff, **kwargs):
    kwargs = fit_kwargs(kwargs)
    x, y, xerr, yerr = fit_split(datax, datay, **kwargs)
    sigmas = yerr
    return stat.Chi2(y, unv(function(x, *ff)), sigmas)


@doc.insert_doc(stat.R2)
def R2(datax, datay, function, ff, **kwargs):
    kwargs = fit_kwargs(kwargs)
    x, y, xerr, yerr = fit_split(datax, datay, **kwargs)
    return stat.R2(y, unv(function(x, *ff)))


def data_split(datax, datay, **kwargs):
    """
    Split data + errors
    """
    return data.__data_split(datax, datay, **kwargs)


def fit_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).

    Parameters
    ----------
    **kwargs : optional
        see :func:`fit_kwargs`.
    """
    kwargs = fit_kwargs(kwargs)
    return data.filtered_data_split(datax, datay, **kwargs)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from smpl import doc
from smpl import util
from smpl import stat
import uncertainties.unumpy as unp
import numpy as np

default = {
    'frange': [None, "Limit the fit to given range. First integer is the lowest and second the highest index.", ],
    'fselector': [None, "Function that takes ``x`` and ``y`` as parameters and returns an array mask in order to limit the data points for fitting. Alternatively a mask for selecting elements from datax and datay.", ],
    'sortbyx': [True, "Enable sorting the x and y data so that x is sorted.", ],
    'bins': [0, "Number of bins for histogram", ],
    'binunc': [stat.poisson_dist, "Number of bins for histogram", ],
    'xerror': [True, "enable xerrors"],
    'yerror': [True, "enable yerrors"],
}


unv = unp.nominal_values
usd = unp.std_devs

# @doc.insert_str("\tDefault kwargs\n\n\t")


@doc.append_str(doc.table(default, init=False))
@doc.append_str(doc.table({"data_kwargs": ["default", "description"]}, bottom=False))
def data_kwargs(kwargs):
    """Set default data_kwargs if not set.

    """
    for k, v in default.items():
        if not k in kwargs:
            kwargs[k] = v[0]
    return kwargs


def __data_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).
    Does not apply filters `frange` and `fselector`.

    Parameters
    ----------
    **kwargs : optional
        see :func:`data_kwargs`.
    """
    if kwargs['bins'] > 0:
        N, bins = np.histogram(unv(datax), bins=kwargs['bins'])
        y = kwargs['binunc'](N)
        yerr = usd(y)
        yerr = yerr if np.any(np.abs(yerr) > 0) else None
        return bins[0:-1] - (bins[0]-bins[1])/2, unv(y), None, yerr
    if util.has("sortbyx", kwargs) and kwargs['sortbyx']:
        ind = np.argsort(unv(datax))
    else:
        ind = np.array(range(len(datax)))
    x = unv(datax)[ind]
    y = unv(datay)[ind]
    xerr = usd(datax)[ind]
    yerr = usd(datay)[ind]
    xerr = xerr if np.any(np.abs(xerr) > 0) else None
    yerr = yerr if np.any(np.abs(yerr) > 0) else None
    if util.has("xerror", kwargs) and not kwargs['xerror']:
        xerr = None
    if util.has("yerror", kwargs) and not kwargs['yerror']:
        yerr = None
    return x, y, xerr, yerr


def _data_split(datax, datay, **kwargs):
    """
    Applies `fselector` and calls :func:`data_split`
    """
    if util.has('fselector', kwargs):
        sel = kwargs['fselector']
        if callable(sel):
            return __data_split(datax[sel(datax, datay)], datay[sel(datax, datay)], **kwargs)
        else:
            return __data_split(datax[sel], datay[sel], **kwargs)
    return __data_split(datax, datay, **kwargs)


def filtered_data_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).
    Applies filters `fselector` and `frange`.

    Returns
    -------
    (x,y,xerr,yerr) : tuple
        four arrays with specified values.

    Parameters
    ----------
    **kwargs : optional
        see :func:`data_kwargs`.
    """
    kwargs = data_kwargs(kwargs)
    x, y, xerr, yerr = _data_split(datax, datay, **kwargs)
    if util.has('frange', kwargs):
        x = x[kwargs['frange'][0]:kwargs['frange'][1]]
        y = y[kwargs['frange'][0]:kwargs['frange'][1]]
        if not yerr is None:
            yerr = yerr[kwargs['frange'][0]:kwargs['frange'][1]]
        if not xerr is None:
            xerr = xerr[kwargs['frange'][0]:kwargs['frange'][1]]

    return x, y, xerr, yerr


def flatmesh(*args):
    """
    Similar to `numpy.meshgrid` but the result will be of one dimensio instead of stacked arrays.
    """
    r = np.meshgrid(*args)
    rr = []
    for ri in r:
        rr += [ri.reshape(ri.size)]
    return rr



"""Simplified statistics."""

from .stat import unv, usd, unv_lambda, poisson_dist, normalize, novar_mean, mean, noisy, no_dist,R2,Chi2

__all__ = ['unv',
           'usd',
           'unv_lambda', 'poisson_dist',
           'no_dist',
           'normalize', 'novar_mean',
           'mean', 'noisy',
           'R2','Chi2'

           ]



import numpy as np


def times(s, n):
    """
    Concats string ``s`` ``n`` times.

    Examples
    --------
    >>> times("hi",5)
    'hihihihihi'

    """
    return s.join(["" for i in range(0, n+1)])


def get(key, ddict, default=None):
    """
    Returns dict[key] if this exists else default.

    Examples
    --------
    >>> d = {'a' : 1 , 'b' : 2 , 'c' : 3}
    >>> get('a',d,5)
    1
    >>> get('x',d,5)
    5

    """
    if has(key, ddict):
        return ddict[key]
    else:
        return default


def has(key, ddict):
    """
    Checks if the key is in the dict and not None.

    Examples
    --------
    >>> d = {'a' : 1 , 'b' : 2 , 'c' : 3}
    >>> has('a',d)
    True
    >>> has('x',d)
    False


    """
    return key in ddict and not ddict[key] is None


def true(key, ddict):
    """
    Checks if the key is in the dict and not None and True.

    Examples
    --------
    >>> d = {'a' : True , 'b' : True , 'c' : False}
    >>> true('a', d)
    True
    >>> true('c', d)
    False
    >>> true('x', d)
    False

    """
    return has(key, ddict) and ddict[key]


def find_nearest_index(array, value):
    """
    Returns the index of the element in ``array`` closest to ``value``

    Examples
    --------
    >>> find_nearest_index([1,7,6,2] , 1.9)
    3

    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


def find_nearest(array, value):
    """
    Return the element in ``array`` closest to ``value``

    Examples
    --------
    >>> find_nearest([1,7,6,2] , 1.9)
    2

    """
    return array[find_nearest_index(array, value)]



import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib as mpl
from PIL import Image
from matplotlib.backends.backend_agg import FigureCanvasAgg
import numpy as np
import ipywidgets

frames = []


class FigAnimation(animation.FuncAnimation):
    def __init__(self, figs=None, frames=None, init=None, update=None,  *args, **kwargs):
        # Use the list of figures as the framedata, which will be iterated
        # over by the machinery.
        if update is None:
            f1 = figs[0]
        else:
            update(frames[0])
            f1 = plt.gcf()
            plt.clf()
            plt.close()

        self._figs = figs
        f = plt.figure()
        plt.axis('off')
        plt.subplots_adjust(top=1, bottom=0, right=1,
                            left=0, hspace=0, wspace=0)
        plt.margins(0, 0)

        canvas = FigureCanvasAgg(f1)
        canvas.draw()
        rgba = np.asarray(canvas.buffer_rgba())
        im = Image.fromarray(rgba)
        self.im = plt.imshow(im)

        def _update(frame):
            ff = None
            if update is not None:
                update(frame)
                ff = plt.gcf()
            else:
                ff = figs[frame]
            canvas = FigureCanvasAgg(ff)
            canvas.draw()
            rgba = np.asarray(canvas.buffer_rgba())
            im = Image.fromarray(rgba)
            self.im.set_array(mpl.image.pil_to_array(im))
            return self.im,

        animation.FuncAnimation.__init__(
            self, f, _update, frames=len(figs) if update is None else frames, init_func=init, *args, **kwargs)


def frame():
    """
    Saves current Matplotlib figure.
    """
    global frames
    # f = plt.gcf()
    # f.figure = f
    # diold = f.canvas.draw_idle
    # f.set_visible = lambda b: vis(f, b)
    # f.canvas.draw_idle = lambda _=None: io.pr(
    #    diold()) if len(frames) == 0 else None

    frames.append(plt.gcf())
    # plt.savefig("test"+str(len(frames))+".jpg")
    plt.close()


def clear():
    """
    Empties stored frames.
    """
    global frames
    frames = []


def animate(**kwargs):
    """
    Make frames to Animation

    Parameters
    ==========

    They are passed directly to ArtistAnimation.

    Returns
    =======
    ArtistAnimation

    """
    global frames
    ani = FigAnimation(frames, **kwargs)
    # clear()
    return ani


def interactive(func, *args, **kwargs):
    ipywidgets.interactive(func, *args, **kwargs)



import pathlib
#from io import StringIO
from smpl import debug
import os
import sys
from pathlib import Path

def read(fname):
    """
    Reads the file ``fname``.
    """
    if(not os.path.exists(fname)):
        return ""
    with open(fname, 'r') as f:
        return f.read()

def write(destination,content):
    """
    Write to file by string or writable :obj:`destiantion`.

    Parameters
    ----------
    destination : str, writeable
        destination to write to.
    content : str
        text to be written.


    Examples
    --------
    >>> write(sys.stdout,"hi")
    hi
    >>> write("test.out","hi")
    """
    # TODO add http and other string based write methodes
    if isinstance(destination,str):
        with open(destination, 'w+') as f:
            f.write(content)
    else:
        destination.write(content)

def gf(i=3):
    """
    Scientific number format.

    Parameters
    ----------
    i : int
        Number of digits.

    Returns
    -------
    str
        Scientific number format string.

    Examples
    --------
    >>> gf(2)
    '{0:.2g}'
    >>> gf(2).format(789234578934)
    '7.9e+11'
    >>> gf(5).format(789234578934)
    '7.8923e+11'

    """
    return "{0:." + str(i) + "g}"


def find_file(fname, up=0):
    """
    Searches for ``fname`` in all down folders or up folder to given order respectively.
    """
    p = Path(fname).parent
    n = Path(fname).name
    if (p / n).is_file():
        return str(p / n)
    for _ in range(up):
        p = p.parent

    for f in p.rglob(n):
        if f.is_file():
            return str(f)
    return None


def pwd():
    """
    Returns the path to the path of current file
    """
    pwd_ = "/".join(debug.get_line_number_file(split=False,
                    _back=1)[1].split("/")[:-1])
    return pwd_


def import_path(path='../..'):
    """
    Adds ``path`` to the ``sys.path``
    """
    sys.path.insert(0, os.path.abspath(path))


def mkdirs(fn):
    """
    Creates the neccessary directories above ``fn``.
    """
    pathlib.Path(fn).parent.mkdir(parents=True, exist_ok=True)


def pr(a, nnl=False):
    """
    Prints the passed ``a``.

    Parameters
    ----------
    nnl : bool
        no-new-line

    Returns
    -------
    a : any
        unchanged ``a``.

    Examples
    --------
    >>> 5 + pr(4)
    4
    9
    >>> 5 + pr(4, nnl=True)
    49

    """
    if nnl:
        print(a, end='')
    else:
        print(a)
    return a


def files(folder, ending):
    """
    Get all the files in ``folder`` ending with ``ending``.
    """
    r = []
    i = 0
    for file in os.scandir(folder):
        if file.path.endswith(ending):
            r.append((i, os.path.splitext(
                os.path.basename(file.path))[0], file.path))
            i = i+1
    return r


def pn(a, nnl=False):
    gl = globals()
    for key in gl:
        if(gl[key] == a):
            print(key)
    if nnl:
        print("%s=%s" % (a.__name__, a), end='')
    else:
        print("%s=%s" % (a.__name__, a))
    return a



import uncertainties as unc
# everything in iminuit is done through the Minuit object, so we import it
from iminuit import Minuit

# we also need a cost function to fit and import the LeastSquares function
from iminuit.cost import LeastSquares
from iminuit.util import Matrix


def _fit_minuit_leastsquares(datax, datay, function, yerr, params=[], **kwargs):
    # TODO check/add params
    least_squares = LeastSquares(datax, datay, yerr, function)
    m = Minuit(least_squares, *params)
    m.migrad()
    m.hesse()
    # print(m.values)
    # print(m.covariance)
    # print("chi2 = ", m.fval)
    # print("ndof = ", len(datax) - m.nfit)

    # fix slice issue from iminuites rewritten __getitem__ by using super
    return unc.correlated_values(m.values, super(Matrix, m.covariance))




import matplotlib.cm as cm
from smpl import doc
from smpl import plot as splot
from matplotlib import colors
import numpy as np
from matplotlib.image import NonUniformImage
import matplotlib.pyplot as plt

default = {
    'xaxis': [None, "."],
    'yaxis': [None, "."],
    'zaxis': [None, "."],
    'logz': [True, "Colorbar in logarithmic scale."],
    'style': ['image', "Plot via an image ('image') or scatter ('scatter')."],
    'interpolation': ['nearest', "Only 'nearest' or 'bilinear' for nonuniformimage. Check https://matplotlib.org/stable/gallery/images_contours_and_fields/interpolation_methods.html#interpolations-for-imshow"],
    'cmap': ['viridis', "Good default color map for missing datapoints since it does not include white."],
    # 'zscale' : [None,"Rescale z values."],
}

# @doc.insert_str("\tDefault kwargs\n\n\t")


@doc.append_str(doc.table(default, init=False))
@doc.append_str(doc.table({"plot2d_kwargs": ["default", "description"]}, bottom=False))
def plot2d_kwargs(kwargs):
    """Set default plot2d_kwargs if not set.

    """
    for k, v in default.items():
        if not k in kwargs:
            kwargs[k] = v[0]
    return kwargs


def plot2d(datax, datay, dataz, **kwargs):
    """
    Creates a 2D-Plot.

    Parameters
    ----------
    **kwargs : optional
        see :func:`plot2d_kwargs`.
    """
    kwargs = plot2d_kwargs(kwargs)
    if kwargs["style"] == "image":
        map_vplot(datax, datay, dataz, **kwargs)
    elif (kwargs["style"] == "scatter"):
        scatter_vplot(datax, datay, dataz, **kwargs)


def map_vplot(tvx,
              tvy,
              tvz,
              xaxis=None,
              yaxis=None,
              zaxis=None,
              logz=True,
              sort=True,
              fill_missing=True,
              zscale=1., **kwargs):
    vx = np.copy(tvx)
    vy = np.copy(tvy)
    vz = np.copy(tvz)
    if fill_missing:
        # TODO speed up
        for x in vx:
            for y in vy:
                ex = np.any(np.logical_and((vx == x), (vy == y)))
                if not ex:
                    vx = np.append(vx, x)
                    vy = np.append(vy, y)
                    vz = np.append(vz, 0)
    if sort:
        p1 = vx.argsort(kind='stable')
        vx = np.copy(vx[p1])
        vy = np.copy(vy[p1])
        vz = np.copy(vz[p1])
        p2 = vy.argsort(kind='stable')
        vx = vx[p2]
        vy = vy[p2]
        vz = vz[p2]
    s = 1
    while vy[s] == vy[s - 1]:
        s = s + 1
    if s == 1:
        #print("flipped x y ")
        while vx[s] == vx[s - 1]:
            s = s + 1
        if s == 1:
            print("error too small map")
            return
        #x, y = y, x
        xaxis, yaxis = yaxis, xaxis
        vx, vy = vy, vx

    grid = splot.unv(vz).reshape((int(np.rint(np.size(vx) / s)), s)) * zscale

    fig, ax = plt.subplots(nrows=1, ncols=1, constrained_layout=True)
    im = None
    xl = vx.min() + (vx.min() / 2) - vx[vx != vx.min()].min() / 2
    xm = vx.max() + (vx.max() / 2) - vx[vx != vx.max()].max() / 2
    yl = vy.min() + (vy.min() / 2) - vy[vy != vy.min()].min() / 2
    ym = vy.max() + (vy.max() / 2) - vy[vy != vy.max()].max() / 2
    im = NonUniformImage(
        ax,
        origin="lower",
        cmap=kwargs['cmap'],
        interpolation=kwargs['interpolation'],
        extent=(xl, xm, yl, ym),
        norm=colors.LogNorm() if logz else None,
    )

    im.set_data(np.unique(vx), np.unique(vy), grid)
    ax.images.append(im)
    ax.set_xlim(xl, xm)
    ax.set_ylim(yl, ym)

    cb = plt.colorbar(im)
    cb.set_label(zaxis)
    plt.xlabel(xaxis)
    plt.ylabel(yaxis)


def scatter_vplot(vx,
                  vy,
                  vz,
                  xaxis=None,
                  yaxis=None,
                  zaxis=None,
                  logz=True,
                  sort=True,
                  fill_missing=True,
                  zscale=1., **kwargs):
    if sort:
        p1 = vx.argsort(kind='stable')
        vx = np.copy(vx[p1])
        vy = np.copy(vy[p1])
        vz = np.copy(vz[p1])
        p2 = vy.argsort(kind='stable')
        vx = vx[p2]
        vy = vy[p2]
        vz = vz[p2]

    fig, ax = plt.subplots(nrows=1, ncols=1, constrained_layout=True)
    im = None
    xl = vx.min() + (vx.min() / 2) - vx[vx != vx.min()].min() / 2
    xm = vx.max() + (vx.max() / 2) - vx[vx != vx.max()].max() / 2
    yl = vy.min() + (vy.min() / 2) - vy[vy != vy.min()].min() / 2
    ym = vy.max() + (vy.max() / 2) - vy[vy != vy.max()].max() / 2

    s = plt.scatter(np.concatenate((vx, vx, vx)),
                    np.concatenate((vy, vy, vy)),
                    c=np.concatenate(
                        (splot.unv(vz) + splot.usd(vz),
                         splot.unv(vz) - splot.usd(vz), splot.unv(vz))),
                    s=np.concatenate(
                        ([(3 * plt.rcParams['lines.markersize'])**2
                          for i in range(len(vx))], [
                              (2 * plt.rcParams['lines.markersize'])**2
                              for i in range(len(vx))
                        ], [(plt.rcParams['lines.markersize'])**2
                            for i in range(len(vx))])),
                    norm=colors.LogNorm() if logz else None,
                    cmap=kwargs['cmap'],)

    ax.set_xlim(xl, xm)
    ax.set_ylim(yl, ym)

    cb = plt.colorbar(s)
    cb.set_label(zaxis)
    plt.xlabel(xaxis)
    plt.ylabel(yaxis)



import numpy as np
import uncertainties as unc
import uncertainties.unumpy as unp
from smpl import doc
import scipy
import math
import statistics as stat
import pandas as pd


unv = unp.nominal_values
usd = unp.std_devs


def R2(y, f):
    """
    R2 - Coefficient of determination

    In the best case, the modeled values exactly match the observed values, which results in R2 = 1. 
    A baseline model, which always predicts the mean of y, will have R2 = 0. 
    Models that have worse predictions than this baseline will have a negative R2. 

    References
    ----------

    https://en.wikipedia.org/wiki/Coefficient_of_determination
    """
    r = y - f
    mean = np.sum(r)/len(r)
    SSres = np.sum((r)**2)
    SStot = np.sum((r-mean)**2)
    Rsq = 1 - SSres/SStot
    return Rsq


def Chi2(y, f, sigmas=None):
    """
    Chi2 - Goodness of Fit

    In general, if Chi-squared/Nd is of order 1.0, then the fit is reasonably good. 
    Coversely,  if Chi-squared/Nd >> 1.0, then the fit is a poor one.

    References
    ----------

    https://www.phys.hawaii.edu/~varner/PHYS305-Spr12/DataFitting.html
    """
    r = y - f
    if sigmas is not None:
        chisq = np.sum((r/sigmas)**2)
    else:
        chisq = np.sum((r)**2)

    return chisq


def unv_lambda(f):
    """Returns a function which applies :func:`unv` on the result of ``f``."""
    return lambda *a: unv(f(*a))


def poisson_dist(N):
    """Return ``N`` with added poissonian uncertainties."""
    return unp.uarray(N, np.sqrt(N))


def no_dist(N):
    """Return ``N`` with no uncertainties."""
    return unp.uarray(N, 0)


def normalize(ydata):
    """Return normalized ``ydata``."""
    return (ydata-np.amin(ydata))/(np.amax(ydata)-np.amin(ydata))


def novar_mean(n):
    """Return mean of ``n`` with only the uncertainties of ``n`` and no variance."""
    return np.sum(n)/len(n)


def mean(n):
    """Return mean of ``n`` with combined error of variance and unvertainties of ``n``."""
    # find the mean value and add uncertainties
    if isinstance(n, pd.core.series.Series):
        n = n.to_numpy()
    k = np.mean(n)
    err = stat.variance(unv(n))
    return unc.ufloat(unv(k), math.sqrt(usd(k)**2 + err))


def noisy(x, mean=0, std=1):
    """Add gaussian noise to ``x``."""
    return x+np.random.normal(mean, std, len(x))


def normal(x, mean=0, std=1):
    return np.random.normal(mean, std, len(x))


@doc.insert_eq()
def fft(y):
    """
    $F(y)$
    """
    N = len(y)
    fft = scipy.fftpack.fft(y)
    return 2 * abs(fft[:N//2]) / N



from tokenize import TokenError
import warnings
from sympy.parsing.sympy_parser import standard_transformations, implicit_multiplication_application
import sympy
from sympy.printing.pycode import pycode
import uncertainties.unumpy as unp
import inspect
import numpy as np


def get_varnames(expr, xvar):
    """
    Returns a list of variables used in the ``str`` math-expression via sympy and puts ``xvar`` to the front.

    Examples
    --------
    >>> get_varnames("a**x*b+c","x")
    ['x', 'a', 'b', 'c']

    """
    if isinstance(expr, str):
        return str_get_varnames(expr, xvar)
    else:
        return fnc_get_varnames(expr, xvar)


def get_latex(function):
    """
    Return a latex string for passed function.


    Parameters
    ----------
    function : function_like
        function as str lambda or (oneline) function

    Examples
    --------
    >>> get_latex(lambda a,b,c,x : (a+b+c)*x,)
    '$x \\\\left(a + b + c\\\\right)$'
    >>> get_latex("(a+b+c)*x")
    '$x \\\\left(a + b + c\\\\right)$'
    >>> def fun(a,b,x,c):
    ...     return (a+b+c)*x
    >>> get_latex(fun)
    '$x \\\\left(a + b + c\\\\right)$'

    """
    if isinstance(function, str):
        l = "$" + sympy.latex(str_get_expr(function)) + "$"
    else:
        l = function.__name__
    if l == "<lambda>":
        try:
            try:
                cc, li = inspect.findsource(function)
                f = ''.join(cc[li:]).split('lambda')[1].split(':')[1].split(',')[
                    0].replace("\n", "")
                l = "$" + sympy.latex(str_get_expr(f)) + "$"
            except TokenError:
                raise Exception(
                    "Make sure there is a ',' behind the lambda expression and no commas are in the lambda expression and no newlines")
        except OSError:
            l = "$\\lambda$(" + ','.join(function.__code__.co_varnames) + ")"
    elif not isinstance(function, str):
        if function.__doc__ is not None:
            l = function.__doc__.split('\n')[0]
        else:
            try:
                cc, li = inspect.findsource(function)
                s = ''.join(cc[li:]).split('return')[1].split('\n')[0]
                l = "$" + sympy.latex(str_get_expr(s)) + "$"
            except OSError:
                l = function.__name__

    return l


def get_lambda_argd(expr, xvar, *args):
    function = get_lambda(expr, xvar)
    return lambda x: function(x, *args)


def get_lambda(expr, xvar):
    """
    Returns a lambda of given ``str``/``function``/``lambda`` expression with ``__doc__`` set to the latex expression. ``xvar`` is moved to the front.

    Examples
    --------
    >>> l = get_lambda(lambda a,b,c,x : (a+b+c)*x,'x')
    >>> l(4,1,1,1)
    12
    >>> l = get_lambda("(a+b+c)*x",'x')
    >>> l(4,1,1,1)
    12
    >>> def fun(a,b,x,c):
    ...     return (a+b+c)*x
    >>> l = get_lambda(fun,'x')
    >>> l(4,1,1,1)
    12

    """
    if isinstance(expr, str):
        return str_get_lambda(expr, xvar)
    else:
        return fnc_get_lambda(expr, xvar)


def fnc_get_lambda(expr, xvar):
    __l__ = eval("lambda " + ','.join(get_varnames(expr, xvar)) +
                             ": expr(" + ','.join(expr.__code__.co_varnames) + ")", {"expr": expr})
    __l__.__doc__ = expr.__doc__
    return __l__


def str_get_lambda(expr, xvar):
    parsed_expr = str_get_expr(expr)
    pc = pycode(parsed_expr)
    for s in unp.__all__:
        pc = pc.replace("math." + s, "unp."+s)
    # old direct way. Doesn't use unp.
    #__l__ = sympy.lambdify(get_varnames(expr),parsed_expr)
    __l__ = eval(
        "lambda " + ','.join(get_varnames(expr, xvar)) + ": " + pc)
    #exec("global __l__; __l__ = lambda " + ','.join(get_varnames(expr)) + ": "+ pc)
    return __l__


def str_get_varnames(expr, xvar):

    parsed_expr = str_get_expr(expr)

    new_locals = [sym.name
                  for sym in parsed_expr.atoms(sympy.Symbol)]
    new_locals = sorted(new_locals)
    if xvar is None:
        return np.roll(new_locals, 1)
    else:
        varss = new_locals
        assert(xvar in varss)
        varss.remove(xvar)
        return [xvar, *varss]


def fnc_get_varnames(func, xvar):
    if xvar is None:
        return func.__code__.co_varnames
    else:
        varss = [s for s in func.__code__.co_varnames]
        assert(xvar in varss)
        varss.remove(xvar)
        return [xvar, *varss]


def str_get_expr(expr):
    """
    Convert a pythonic string expression ot a sympy expression.

    Only works with np or unp naming.
    """
    expr = expr.replace("math.abs(", "Abs(").replace(
        "np.abs(", "Abs(").replace("unp.abs(", "Abs(")
    expr = expr.replace("math.", "").replace("unp.", "").replace("np.", "")
    try:
        parsed_expr = sympy.parsing.sympy_parser.parse_expr(
            expr,
            local_dict=None,
            transformations=(standard_transformations +
                             (implicit_multiplication_application,)),
            evaluate=False
        )
    except Exception as e:
        warnings.warn(
            "Illegal variable/function name (try uncap. letters) " + expr)
        raise e
    return parsed_expr



"""
Simplified Fitting.

Uses scipy.curve_fit (no x errors) or scipy.odr (with x errors).
"""
from .fit import Fitter, fit, auto, fit_kwargs, data_split, fit_split, Chi2, R2
__all__ = ['Fitter', 'fit', 'auto', 'fit_kwargs',
           'data_split', 'fit_split', 'Chi2', 'R2']



import warnings
from scipy.interpolate import make_interp_spline, BSpline
from scipy import interpolate as interp
from smpl import doc
from smpl import plot as splot
from smpl import data
import numpy as np
import uncertainties as unc

default = {
    'interpolator': ['cubic', "Use 'cubic' (spline) or 'linear' (cf. <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>)."],
    'order': [3, "Spline order."],
    'pre': [lambda x:x, "pre interpolation"],
    'post': [lambda x:x, "post interpolation"],
    'interpolate_upper_uncertainty': [True, "Interpolate only using upper uncertainty."],
    'interpolate_lower_uncertainty': [True, "Interpolate only using lower uncertainty."],
}

# @doc.insert_str("\tDefault kwargs\n\n\t")


@doc.append_doc(data.data_kwargs)
@doc.append_str("\t")
@doc.append_str(doc.table(default, init=False))
@doc.append_str(doc.table({"interpolate_kwargs": ["default", "description"]}, bottom=False))
def interpolate_kwargs(kwargs):
    """Set default interpolate_kwargs if not set.

    """
    kwargs = data.data_kwargs(kwargs)
    for k, v in default.items():
        if not k in kwargs:
            kwargs[k] = v[0]
    return kwargs


def interpolate_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).

    Parameters
    ----------
    **kwargs : optional
        see :func:`interpolate_kwargs`.
    """
    kwargs = interpolate_kwargs(kwargs)
    return data.filtered_data_split(datax, datay, **kwargs)


def interpolate(*data, **kwargs):
    # TODO save spl_upd spl_down values instead of recalculate
    # TODO set eps
    kwargs = interpolate_kwargs(kwargs)
    kwargs['sortbyx'] = False
    x, y, dx, dy = interpolate_split(data[0], data[-1], **kwargs)
    ret = None
    if dy is None:
        spl_center = _interpolate(*data[:-1], kwargs['pre'](y), **kwargs)
        ret = np.vectorize(lambda *a: spl_center(*a), otypes=["float"])
    else:
        if kwargs['interpolate_upper_uncertainty'] and kwargs['interpolate_lower_uncertainty']:
            spl_up = _interpolate(*data[:-1], kwargs['pre'](y+dy), **kwargs)
            spl_down = _interpolate(*data[:-1], kwargs['pre'](y-dy), **kwargs)
            ret = np.vectorize(lambda *a: unc.ufloat(spl_up(*a)/2 + spl_down(*a)/2,
                                                 np.abs(spl_up(*a) - spl_down(*a))/2), otypes=["object"])  # symmetrized error...
        elif not kwargs['interpolate_lower_uncertainty']:
            spl_center = _interpolate(*data[:-1], kwargs['pre'](y), **kwargs)
            spl_up = _interpolate(*data[:-1], kwargs['pre'](y+dy), **kwargs)
            ret = np.vectorize(lambda *a: unc.ufloat(spl_center(*a),
                                                 np.abs(spl_up(*a) - spl_center(*a))), otypes=["object"])  # symmetrized error...
        elif not kwargs['interpolate_upper_uncertainty']:
            spl_center = _interpolate(*data[:-1], kwargs['pre'](y), **kwargs)
            spl_down = _interpolate(*data[:-1], kwargs['pre'](y-dy), **kwargs)
            ret = np.vectorize(lambda *a: unc.ufloat(spl_center(*a),
                                                 np.abs(spl_down(*a) - spl_center(*a))), otypes=["object"])  # symmetrized error...
        else:
            raise ValueError("interpolate_upper_uncertainty and interpolate_lower_uncertainty can't be both False")
            spl_center = _interpolate(*data[:-1], kwargs['pre'](y), **kwargs)
            ret = np.vectorize(lambda *a: spl_center(*a), otypes=["float"])

    if not check(ret, *data):
        warnings.warn("Bad interpolation. Increase Order.")
    return ret


def check(f, *args):
    return np.all(np.isclose(splot.unv(f(*args[:-1])), splot.unv(args[-1]), rtol=1e-2))


def _interpolate(*data, **kwargs):
    ret = None
    if len(data) == 2:
        if kwargs['interpolator'] == 'exp':
            ret = _interpolate_exp(*data)
        else:
            ret = interp.interp1d(*data, kind=kwargs['interpolator'])
    elif len(data) == 3:
        if kwargs['interpolator'] == 'bivariatespline':
            ret = interp.SmoothBivariateSpline(
                *data, kx=kwargs['order'], ky=kwargs['order'])
        elif kwargs['interpolator'] == 'linearnd':
            ret = interp.LinearNDInterpolator(list(zip(*data[:-1])), data[-1])
        else:
            ret = interp.interp2d(*data, kind=kwargs['interpolator'])
    else:
        if kwargs['interpolator'] == 'linear':
            ret = interp.LinearNDInterpolator(list(zip(*data[:-1])), data[-1])

    return lambda *a: kwargs['post'](ret(*a))


def _interpolate_exp(x, y, **kwargs):
    ip = interp.interp1d(x, np.log(y), kind="linear")
    return lambda x_new: np.exp(ip(x_new))



"""A collection of simplified utilities."""

import pkg_resources as pkg  # part of setuptools

package = "smpl"

try:
    version = pkg.require(package)[0].version
except pkg.DistributionNotFound:
    version = "dirty"

__version__ = version


import warnings
import numpy as np
from scipy import optimize
from scipy.odr.odrpack import ODR, Model, RealData
import uncertainties as unc

from smpl import debug
from smpl.util import util
# fittet ein dataset mit gegebenen x und y werten, eine funktion und ggf. anfangswerten und y-Fehler
# gibt die passenden parameter der funktion, sowie dessen unsicherheiten zurueck
#
# https://stackoverflow.com/questionsquestions/14581358/getting-standard-errors-on-fitted-parameters-using-the-optimize-leastsq-method-i#
# Updated on 4/6/2016
# User: https://stackoverflow.com/users/1476240/pedro-m-duarte


def _fit_curvefit(datax, datay, function, params=None, yerr=None, **kwargs):
    try:
        pfit, pcov = \
            optimize.curve_fit(function, datax, datay, p0=params,
                               sigma=yerr, epsfcn=util.get("epsfcn", kwargs, 0.0001), **kwargs, maxfev=util.get("maxfev", kwargs, 10000))
    except RuntimeError as e:
        debug.msg(str(e))
        return params
    error = []
    for i in range(len(pfit)):
        try:
            error.append(np.absolute(pcov[i][i])**0.5)
        except Exception as e:
            warnings.warn(str(e))
            error.append(0.00)
    # print(pcov)
    # restore cov: unc.covariance_matrix([*ff])
    return unc.correlated_values(pfit, pcov)

# https://stackoverflow.com/a/52592811


def _fit_odr(datax, datay, function, params=None, yerr=None, xerr=None):
    model = Model(lambda p, x: function(x, *p))
    realdata = RealData(datax, datay, sy=yerr, sx=xerr)
    odr = ODR(realdata, model, beta0=params)
    out = odr.run()
    # This was the old wrong way! Now use correct co. matrix through unc-package!
    # Note Issues on scipy odr and curve_fit, regarding different definitions/namings of standard deviation or error and covaraince matrix
    # https://github.com/scipy/scipy/issues/6842
    # https://github.com/scipy/scipy/pull/12207
    # https://stackoverflow.com/questions/62460399/comparison-of-curve-fit-and-scipy-odr-absolute-sigma
    return unc.correlated_values(out.beta, out.cov_beta)



from inspect import currentframe, getsource
import os
import numpy as np

DEBUG_LEVEL = -1
# DEBUG_LEVEL=-1 no debug
# DEBUG_LEVEL= 0 default level
# DEBUG_LEVEL= 1,2,3 higher levels, more debug

# print prefix
DEBUG_PRE = "DBG"

# remove folders of files for print

# filename strings with .py (with folder if SPLIT_FILENAME==False)
WHITE_LIST_FILES = []
BLACK_LIST_FILES = []

active = debug = on = DEBUG_LEVEL >= 0
# count debug events by file+line
count_times = {}
cur_table_line = {}


def get_frame(_back=0):
    cf = currentframe()
    for _ in range(_back+1):
        cf = cf.f_back
    return cf


def once(_back=0):
    """
    Returns true only one time

    Examples
    --------
    >>> for i in range(10):
    ...     if once():
    ...         print(i)
    0

    """
    return times(1, _back+1)


def times(t=1, _back=0):
    line, fname = get_line_number_file(_back+1)
    inc_count(line, fname)
    return check_count(line, fname, t)


def get_line_src(_back=0):
    cf = get_frame(_back+1)
    srcline = getsource(cf).split(
        "\n")[cf.f_lineno-cf.f_code.co_firstlineno].strip()
    return srcline


def get_line_number_file(split=True, _back=0):
    """
    Gets the current filename and the current linenumber within it.

    Parameters
    ----------
    split : bool
        Indicates whenever the folders above of the file should be included in the returned filename.
    _back : int
        Number of stack/frames to go back.

    Returns
    -------
    filenumber : int
        First element in the return array
    filename : str
        Second element in the return array

    """
    cf = get_frame(_back+1)
    fname = cf.f_code.co_filename
    if split:
        fname = cf.f_code.co_filename.split("/")[-1]
    return cf.f_lineno, fname


def get_line_number(_back=0):
    return get_line_number_file(_back+1)[0]


def line(msg_, tag="", level=0, times=-1, _back=0, **kwargs):
    msg(msg_, tag=tag, level=level, times=times,
        line_=True, _back=_back+1, **kwargs)
# only once


def line1(msg_, tag="", level=0, times=-1, _back=0, **kwargs):
    msg1(msg_, tag=tag, level=level, times=times,
         line_=True, _back=_back+1, **kwargs)

# counting functions


def get_count(line, fname):
    global count_times
    return count_times[fname+str(line)]


def inc_count(line, fname):
    global count_times
    if fname+str(line) in count_times:
        count_times[fname+str(line)] += 1
    else:
        count_times[fname+str(line)] = 1


def check_count(line, fname, t):
    if t >= get_count(line, fname) or t == -1:
        if(not fname in BLACK_LIST_FILES and (len(WHITE_LIST_FILES) == 0 or fname in WHITE_LIST_FILES)):
            return True
    return False


# _line enables printint src line
# t stands for times
def msg(msg, tag="", level=0, times=-1, line_=False, _back=0, **kwargs):
    """
    Prints the message ``msg`` if level > debug_level and always returns the msg.

    Parameters
    ----------
    tag : str
        Sets a tag to be printed for the debug message.
    level : int
        Debug level.
    times : int
        How often should the message be printed if the function gets called multiple times (e.g. in a loop).
    _line : bool
        Print the current line in the python source.
    _back : int
        Number of stack/frames to go back.

    Examples
    --------
    >>> msg("hi", level = -9999)
    DBG::debug.py:...: hi
    'hi'
    >>> msg("hi")
    'hi'

    """
    if(level <= DEBUG_LEVEL):
        line, fname = get_line_number_file(_back+1)
        src = ""
        if line_ == True:
            src = get_line_src(_back+1)
            src = "(" + '('.join(src.split("(")[1:]) + " = "
        inc_count(line, fname)
        if(check_count(line, fname, times)):
            print(DEBUG_PRE + ":" + tag + ":" + fname +
                  ":" + str(line) + ": " + src + str(msg))
    return msg


def msg1(_msg, tag="", level=0, times=1, line_=False, _back=0, **kwargs):
    """
    Just like :func:`msg` but ``times`` set to 1.
    """
    return msg(_msg, level=level, tag=tag, times=times, line_=line_, _back=_back+1, **kwargs)


def file(key, value, level=0, times=-1, seperator=";", _print=True, _back=0, filename="debug.csv"):
    """
    Prints the message ``msg`` if level > debug_level to file ``filename``
    """
    if(level <= DEBUG_LEVEL):
        line, fname = get_line_number_file(_back+1)
        inc_count(line, fname)
        if(check_count(line, fname, times)):
            f = open(filename, "a+")
            if(_print):
                tag = "debug.file"
                print(DEBUG_PRE + ":" + tag + ":" + fname + ":" +
                      str(line) + ": " + key + seperator + value)
            f.write(key + seperator + value + "\n")
            f.close()
    return value


def file1(_key, _value, level=0, times=1, _back=0, **kwargs):
    """
    Just like :func:`file` but ``times`` set to 1.
    """
    return file(_key, _value, level=level, times=times, _back=_back+1, **kwargs)


def table_flush_header(filename="debug_table.csv", seperator=";"):
    """
    Saves the current keys from :func:`table` to ``filename``
    """
    global cur_table_line
    f = open(filename, "a+")
    for key in sorted(cur_table_line):
        f.write(key + seperator)
    f.write('\n')
    f.close()


def table_flush_line(filename="debug_table.csv", seperator=";"):
    """
    Saves the current values from :func:`table` to ``filename``
    """
    global cur_table_line
    f = open(filename, "a+")
    itt = iter(cur_table_line)
    ok = False
    while not ok:
        cur = next(itt)
        ok = isinstance(cur_table_line[cur], np.ndarray)
    dim = len(cur_table_line[cur])
    for i in range(dim):
        for key in sorted(cur_table_line):
            if isinstance(cur_table_line[key], np.ndarray):
                f.write("%.30e" % cur_table_line[key][i] + seperator)
            else:
                f.write("%.30e" % cur_table_line[key] + seperator)
        f.write('\n')
    f.close()


def table(key, value, level=0, times=-1, seperator=";", _print=False, _back=0, filename="debug_table.csv"):
    """
    Saves ``key``:``value`` in ``filename``.
    """
    global cur_table_line
    if(level <= DEBUG_LEVEL):
        line, fname = get_line_number_file(_back+1)
        inc_count(line, fname)
        if(check_count(line, fname, times)):
            if isinstance(value, np.ndarray):
                cur_table_line[key] = value.copy()
            else:
                cur_table_line[key] = value
    return value


def reset_times():
    """
    Resets global `count_times`.
    """
    global count_times
    count_times = {}


if os.path.exists("debug.csv"):
    os.remove("debug.csv")
if os.path.exists("debug_table.csv"):
    os.remove("debug_table.csv")

if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="smpl",
    setup_requires=['setuptools-git-versioning'],
    author="APN",
    author_email="APN-Pucky@no-reply.github.com",
    description="simple plotting and fitting",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/APN-Pucky/smpl",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "uncertainties",
        "numpy",
        "matplotlib",
        "scipy",
        "sympy",
        "tqdm",
        "pandas",
        "ipywidgets",
        "iminuit",
        # "requests",
    ],
    extras_require={
        'dev': [
            "build",
            "pytest",
            "pytest-cov",
            "jupyterlab",
            "pandas",
            "ipython",
        ],
        'docs': [
            "jupyter-sphinx",
            "sphinx-math-dollar",
            "pandoc",
            "sphinx",
            "nbsphinx",
            "sphinx-rtd-theme",
            "numpydoc",
        ]
    },
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.{ccount}",
        "dirty_template": "{tag}.{ccount}+dirty",
        "starting_version": "0.0.0",
        "version_callback": None,
        "version_file": None,
        "count_commits_from_version_file": False
    },
    python_requires='>=3.6',
)



"""Simplified internal data parsing and transforming"""

from .data import filtered_data_split, __data_split, data_kwargs, flatmesh
__all__ = ['filtered_data_split', '__data_split', 'data_kwargs', 'flatmesh']



"""Simplified plotting."""

from .plot import data, fit, function, unv, usd, auto, plot_kwargs
from .plot2d import plot2d, plot2d_kwargs
__all__ = ['data', 'fit', 'function', 'unv', 'usd',
           'auto', 'plot_kwargs', 'plot2d', 'plot2d_kwargs']
from matplotlib.pyplot import *


"""A list of various functions."""
from .functions import const, lorentz, linear, line, cos_abs, gauss, square, cube, exp, log, sqrt, tan, cos, sin

__all__ = [
    'const',
    'lorentz',
    'cos',
    'sin',
    'tan',
    'linear',
    'line',
    'cos_abs',
    'gauss',
    'square',
    'cube',
    'sqrt',
    'exp',
    'log'
]
